@@ -127,6 +127,20 @@ fn test_horizontal_rule() {
     assert_eq!(result, "---\n");
 }
 
+#[test]
+fn test_hr_after_paragraph_keeps_blank_line() {
+    let html = "<p>paragraph</p><hr>";
+    let result = convert(html, None).unwrap();
+    assert_eq!(result, "paragraph\n\n---\n");
+}
+
+#[test]
+fn test_comment_between_paragraphs() {
+    let html = "<p>yes</p><!----><p>but no</p>";
+    let result = convert(html, None).unwrap();
+    assert_eq!(result, "yes\n\nbut no\n");
+}
+
 #[test]
 fn test_line_break() {
     let html = "<p>Line 1<br>Line 2</p>";
